/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef XUNFEISPEECHENGINE_P_H
#define XUNFEISPEECHENGINE_P_H

#include <thread>
#include <cpr/cpr.h>
#include <curl/curl.h>
#include <jsoncpp/json/json.h>
#include <kylin-ai/ai-engine/cloudspeechengine.h>

// 开始：ReconnectSettings
class ReconnectSettings {
public:
    using Callback = std::function<void()>;

public:
    void setCallback(const Callback &callback) { callback_ = callback; }

    void setLimitRetryCount(int32_t count) { limitRetryCount_ = count; }

    void setRetryIntervalMs(int32_t ms) { retryIntervalMs_ = ms; }

    /**
     * 超过重试次数，不重试
     * 第一次运行，不重试
     */
    void retryOnce();

    void resetRetryParam();

    bool canRetry() const;

    bool isRetrying() const { return retrying_; }

private:
    void callCallback() const;

    void initRetryParam();

private:
    Callback callback_;
    int32_t limitRetryCount_{5}; // 限制重试次数
    int32_t retryIntervalMs_{1000};
    int64_t currentTimestamp_{0};
    std::atomic<int32_t> currentRetryCount_{0}; // 当前重试次数
    std::atomic<bool> retrying_{false};
};
// 结束：ReconnectSettings

class XunfeiSpeechEnginePrivate
    : public ai_engine::speech::AbstractCloudSpeechEngine {
public:
    XunfeiSpeechEnginePrivate();
    ~XunfeiSpeechEnginePrivate() override;
    std::string engineName() const override;
    std::string currentModelName() const override;
    int maxConcurrentOnceRecognitionTasks() const override;
    int maxConcurrentContinuousRecognitionTasks() const override;
    int maxConcurrentOnceSynthesisTasks() const override;
    int maxConcurrentContinuousSynthesisTasks() const override;

    void setConfig(const std::string &config) override;
    std::string modelInfo() const override;

    void setRecognizingCallback(
        ai_engine::speech::RecognitionResultCallback callback) override;
    void setRecognizedCallback(
        ai_engine::speech::RecognitionResultCallback callback) override;

    void setSynthesizingCallback(
        ai_engine::speech::SynthesisResultCallback callback) override;
    void setSynthesizedCallback(
        ai_engine::speech::FinishedCallback callback) override;

    // 语音识别
    bool
    initContinuousRecognitionModule(ai_engine::EngineError &error) override;
    bool startContinuousRecognition(const std::string &params,
                                    ai_engine::EngineError &error) override;
    bool
    writeContinuousRecognitionAudioData(const std::vector<char> &audioData,
                                        ai_engine::EngineError &error) override;
    bool stopContinuousRecognition(ai_engine::EngineError &error) override;
    bool
    destroyContinuousRecognitionModule(ai_engine::EngineError &error) override;

    bool initRecognizeOnceModule(ai_engine::EngineError &error) override;
    bool recognizeOnce(const std::string &params,
                       const std::vector<char> &audioData,
                       ai_engine::EngineError &error) override;
    bool destroyRecognizeOnceModule(ai_engine::EngineError &error) override;

    // 语音合成
    bool initContinuousSynthesisModule(ai_engine::EngineError &error) override;
    bool startContinuousSynthesis(const std::string &params,
                                  ai_engine::EngineError &error) override;
    bool writeContinuousSynthesisText(const std::string &text,
                                      ai_engine::EngineError &error) override;
    bool stopContinuousSynthesis(ai_engine::EngineError &error) override;
    bool
    destroyContinuousSynthesisModule(ai_engine::EngineError &error) override;

    bool initSynthesizeOnceModule(ai_engine::EngineError &error) override;
    bool synthesizeOnce(const std::string &params, const std::string &text,
                        ai_engine::EngineError &error) override;
    bool destroySynthesizeOnceModule(ai_engine::EngineError &error) override;

private:
    using RecognitionResultList =
        std::vector<ai_engine::speech::RecognitionResult>;
    void handleRecognitionResult(const Json::Value &jsonData);
    bool isRecognizingResult(const Json::Value &jsonData);

    RecognitionResultList
    getContinuousRecognitionResult(const Json::Value &jsonData);

    // 流式语音识别的重连
    void continuousAsrReconnectWebSockets();

private:
    enum class OperationType {
        RecognizeOnce,
        ContinuousRecognition,
        SynthesizeOnce,
        ContinuousSynthesis
    };

    const char *asrHostUrl_ = "wss://iat-api.xfyun.cn/v2/iat";
    const char *realtimeAsrHostUrl_ = "wss://rtasr.xfyun.cn/v1/ws";
    const char *ttsHostUrl_ = "wss://tts-api.xfyun.cn/v2/tts";

    std::string apiKeyRealTime_;
    std::string appIdRealTime_;
    std::string apiKeyContinuous_;
    std::string secretKeyContinuous_;
    std::string appIdContinuous_;

    ai_engine::speech::RecognitionResultCallback recognizingCallback_;
    ai_engine::speech::RecognitionResultCallback recognizedCallback_;
    ai_engine::speech::SynthesisResultCallback synthesizingCallback_;
    ai_engine::speech::FinishedCallback synthesizedCallback_;

    bool doSendData(const char *data, int dataLength, int type);

    bool doSendEndFlag();

    std::string format_;
    int rate_{0};
    int channel_{0};
    int currentRole_{1};
    void setRecognitionParams_(const std::string &params);
    bool doReceiveOnceRecognitionData();
    friend bool
    doReceiveContinuousRecognitionData(XunfeiSpeechEnginePrivate *self);

    int speed_{50};
    int volume_{50};
    int pitch_{50};
    Json::Value lastRecognitionResult_;
    OperationType currentOperation_;
    void setSynthesisParams_(const std::string &params);
    friend bool doReceiveSynthesisData(XunfeiSpeechEnginePrivate *self);

    const int maxWebDelay_{1000 * 10};
    int currentWebDelay_{0};
    bool isPinging_{false};
    bool connectWebsocket(const std::string &hostUrl);
    bool connectRealtimeAsrWebsocket();
    friend void
    onWebSocket(bool (*doReceiveData)(XunfeiSpeechEnginePrivate *self),
                XunfeiSpeechEnginePrivate *self);
    bool testWebSocket();
    void closeWebsocket();

    CURL *currentCurl_{nullptr};
    std::thread *currentThread_{nullptr};
    ai_engine::EngineError currentError_;
    bool currentStatus();
    void runCallbackWithError(const ai_engine::EngineError &error,
                              OperationType type);

    void releaseResources();

    std::string incompleteResult_;

    bool isStopped_{false};

    ReconnectSettings reconnectSettings_;
};

#endif
